@@ -4,10 +4,12 @@ Streamline plotting for 2D vector fields.
 """
 
 import numpy as np
+
 import matplotlib
 import matplotlib.cm as cm
 import matplotlib.colors as mcolors
 import matplotlib.collections as mcollections
+import matplotlib.lines as mlines
 import matplotlib.patches as patches
 
 
@@ -16,64 +18,80 @@ __all__ = ['streamplot']
 
 def streamplot(axes, x, y, u, v, density=1, linewidth=None, color=None,
                cmap=None, norm=None, arrowsize=1, arrowstyle='-|>',
-               minlength=0.1, transform=None):
-    """Draws streamlines of a vector flow.
+               minlength=0.1, transform=None, zorder=None, start_points=None,
+               maxlength=4.0, integration_direction='both'):
+    """
+    Draw streamlines of a vector flow.
 
-    *x*, *y* : 1d arrays
-        an *evenly spaced* grid.
-    *u*, *v* : 2d arrays
-        x and y-velocities. Number of rows should match length of y, and
-        the number of columns should match x.
-    *density* : float or 2-tuple
-        Controls the closeness of streamlines. When `density = 1`, the domain
-        is divided into a 25x25 grid---*density* linearly scales this grid.
+    Parameters
+    ----------
+    x, y : 1d arrays
+        An evenly spaced grid.
+    u, v : 2d arrays
+        *x* and *y*-velocities. Number of rows should match length of *y*, and
+        the number of columns should match *x*.
+    density : float or 2-tuple
+        Controls the closeness of streamlines. When ``density = 1``, the domain
+        is divided into a 30x30 grid---*density* linearly scales this grid.
         Each cell in the grid can have, at most, one traversing streamline.
         For different densities in each direction, use [density_x, density_y].
-    *linewidth* : numeric or 2d array
-        vary linewidth when given a 2d array with the same shape as velocities.
-    *color* : matplotlib color code, or 2d array
+    linewidth : numeric or 2d array
+        Vary linewidth when given a 2d array with the same shape as velocities.
+    color : matplotlib color code, or 2d array
         Streamline color. When given an array with the same shape as
         velocities, *color* values are converted to colors using *cmap*.
-    *cmap* : :class:`~matplotlib.colors.Colormap`
+    cmap : `~matplotlib.colors.Colormap`
         Colormap used to plot streamlines and arrows. Only necessary when using
         an array input for *color*.
-    *norm* : :class:`~matplotlib.colors.Normalize`
-        Normalize object used to scale luminance data to 0, 1. If None, stretch
-        (min, max) to (0, 1). Only necessary when *color* is an array.
-    *arrowsize* : float
+    norm : `~matplotlib.colors.Normalize`
+        Normalize object used to scale luminance data to 0, 1. If ``None``,
+        stretch (min, max) to (0, 1). Only necessary when *color* is an array.
+    arrowsize : float
         Factor scale arrow size.
-    *arrowstyle* : str
+    arrowstyle : str
         Arrow style specification.
-        See :class:`~matplotlib.patches.FancyArrowPatch`.
-    *minlength* : float
+        See `~matplotlib.patches.FancyArrowPatch`.
+    minlength : float
         Minimum length of streamline in axes coordinates.
+    start_points : Nx2 array
+        Coordinates of starting points for the streamlines.
+        In data coordinates, the same as the *x* and *y* arrays.
+    zorder : int
+        Any number.
+    maxlength : float
+        Maximum length of streamline in axes coordinates.
+    integration_direction : ['forward' | 'backward' | 'both']
+        Integrate the streamline in forward, backward or both directions.
+        default is ``'both'``.
 
-    Returns:
+    Returns
+    -------
+    stream_container : StreamplotSet
+        Container object with attributes
 
-        *stream_container* : StreamplotSet
-            Container object with attributes
+        - lines: `matplotlib.collections.LineCollection` of streamlines
 
-                - lines: `matplotlib.collections.LineCollection` of streamlines
-
-                - arrows: collection of `matplotlib.patches.FancyArrowPatch`
-                  objects representing arrows half-way along stream
-                  lines.
-
-            This container will probably change in the future to allow changes
-            to the colormap, alpha, etc. for both lines and arrows, but these
-            changes should be backward compatible.
+        - arrows: collection of `matplotlib.patches.FancyArrowPatch`
+          objects representing arrows half-way along stream
+          lines.
 
+        This container will probably change in the future to allow changes
+        to the colormap, alpha, etc. for both lines and arrows, but these
+        changes should be backward compatible.
     """
     grid = Grid(x, y)
     mask = StreamMask(density)
     dmap = DomainMap(grid, mask)
 
+    if zorder is None:
+        zorder = mlines.Line2D.zorder
+
     # default to data coordinates
     if transform is None:
         transform = axes.transData
 
     if color is None:
-        color = next(axes._get_lines.color_cycle)
+        color = axes._get_lines.get_next_color()
 
     if linewidth is None:
         linewidth = matplotlib.rcParams['lines.linewidth']
@@ -81,38 +99,74 @@ def streamplot(axes, x, y, u, v, density=1, linewidth=None, color=None,
     line_kw = {}
     arrow_kw = dict(arrowstyle=arrowstyle, mutation_scale=10 * arrowsize)
 
+    if integration_direction not in ['both', 'forward', 'backward']:
+        errstr = ("Integration direction '%s' not recognised. "
+                  "Expected 'both', 'forward' or 'backward'." %
+                  integration_direction)
+        raise ValueError(errstr)
+
+    if integration_direction == 'both':
+        maxlength /= 2.
+
     use_multicolor_lines = isinstance(color, np.ndarray)
     if use_multicolor_lines:
-        assert color.shape == grid.shape
+        if color.shape != grid.shape:
+            raise ValueError(
+                "If 'color' is given, must have the shape of 'Grid(x,y)'")
         line_colors = []
-        if np.any(np.isnan(color)):
-            color = np.ma.array(color, mask=np.isnan(color))
+        color = np.ma.masked_invalid(color)
     else:
         line_kw['color'] = color
         arrow_kw['color'] = color
 
     if isinstance(linewidth, np.ndarray):
-        assert linewidth.shape == grid.shape
+        if linewidth.shape != grid.shape:
+            raise ValueError(
+                "If 'linewidth' is given, must have the shape of 'Grid(x,y)'")
         line_kw['linewidth'] = []
     else:
         line_kw['linewidth'] = linewidth
         arrow_kw['linewidth'] = linewidth
 
+    line_kw['zorder'] = zorder
+    arrow_kw['zorder'] = zorder
+
     ## Sanity checks.
-    assert u.shape == grid.shape
-    assert v.shape == grid.shape
+    if u.shape != grid.shape or v.shape != grid.shape:
+        raise ValueError("'u' and 'v' must be of shape 'Grid(x,y)'")
 
-    if np.any(np.isnan(u)):
-        u = np.ma.array(u, mask=np.isnan(u))
-    if np.any(np.isnan(v)):
-        v = np.ma.array(v, mask=np.isnan(v))
+    u = np.ma.masked_invalid(u)
+    v = np.ma.masked_invalid(v)
 
-    integrate = get_integrator(u, v, dmap, minlength)
+    integrate = get_integrator(u, v, dmap, minlength, maxlength,
+                               integration_direction)
 
     trajectories = []
-    for xm, ym in _gen_starting_points(mask.shape):
-        if mask[ym, xm] == 0:
-            xg, yg = dmap.mask2grid(xm, ym)
+    if start_points is None:
+        for xm, ym in _gen_starting_points(mask.shape):
+            if mask[ym, xm] == 0:
+                xg, yg = dmap.mask2grid(xm, ym)
+                t = integrate(xg, yg)
+                if t is not None:
+                    trajectories.append(t)
+    else:
+        sp2 = np.asanyarray(start_points, dtype=float).copy()
+
+        # Check if start_points are outside the data boundaries
+        for xs, ys in sp2:
+            if not (grid.x_origin <= xs <= grid.x_origin + grid.width
+                    and grid.y_origin <= ys <= grid.y_origin + grid.height):
+                raise ValueError("Starting point ({}, {}) outside of data "
+                                 "boundaries".format(xs, ys))
+
+        # Convert start_points from data to array coords
+        # Shift the seed points from the bottom left of the data so that
+        # data2grid works properly.
+        sp2[:, 0] -= grid.x_origin
+        sp2[:, 1] -= grid.y_origin
+
+        for xs, ys in sp2:
+            xg, yg = dmap.data2grid(xs, ys)
             t = integrate(xg, yg)
             if t is not None:
                 trajectories.append(t)
@@ -131,14 +185,15 @@ def streamplot(axes, x, y, u, v, density=1, linewidth=None, color=None,
         tgx = np.array(t[0])
         tgy = np.array(t[1])
         # Rescale from grid-coordinates to data-coordinates.
-        tx = np.array(t[0]) * grid.dx + grid.x_origin
-        ty = np.array(t[1]) * grid.dy + grid.y_origin
+        tx, ty = dmap.grid2data(*np.array(t))
+        tx += grid.x_origin
+        ty += grid.y_origin
 
         points = np.transpose([tx, ty]).reshape(-1, 1, 2)
         streamlines.extend(np.hstack([points[:-1], points[1:]]))
 
         # Add arrows half way along each trajectory.
-        s = np.cumsum(np.sqrt(np.diff(tx) ** 2 + np.diff(ty) ** 2))
+        s = np.cumsum(np.hypot(np.diff(tx), np.diff(ty)))
         n = np.searchsorted(s, s[-1] / 2.)
         arrow_tail = (tx[n], ty[n])
         arrow_head = (np.mean(tx[n:n + 2]), np.mean(ty[n:n + 2]))
@@ -150,27 +205,24 @@ def streamplot(axes, x, y, u, v, density=1, linewidth=None, color=None,
 
         if use_multicolor_lines:
             color_values = interpgrid(color, tgx, tgy)[:-1]
-            line_colors.extend(color_values)
+            line_colors.append(color_values)
             arrow_kw['color'] = cmap(norm(color_values[n]))
 
-        p = patches.FancyArrowPatch(arrow_tail,
-                                    arrow_head,
-                                    transform=transform,
-                                    **arrow_kw)
+        p = patches.FancyArrowPatch(
+            arrow_tail, arrow_head, transform=transform, **arrow_kw)
         axes.add_patch(p)
         arrows.append(p)
 
-    lc = mcollections.LineCollection(streamlines,
-                                     transform=transform,
-                                     **line_kw)
+    lc = mcollections.LineCollection(
+        streamlines, transform=transform, **line_kw)
+    lc.sticky_edges.x[:] = [grid.x_origin, grid.x_origin + grid.width]
+    lc.sticky_edges.y[:] = [grid.y_origin, grid.y_origin + grid.height]
     if use_multicolor_lines:
-        lc.set_array(np.asarray(line_colors))
+        lc.set_array(np.ma.hstack(line_colors))
         lc.set_cmap(cmap)
         lc.set_norm(norm)
     axes.add_collection(lc)
-
-    axes.update_datalim(((x.min(), y.min()), (x.max(), y.max())))
-    axes.autoscale_view(tight=True)
+    axes.autoscale_view()
 
     ac = matplotlib.collections.PatchCollection(arrows)
     stream_container = StreamplotSet(lc, ac)
@@ -185,7 +237,7 @@ class StreamplotSet(object):
 
 
 # Coordinate definitions
-#========================
+# ========================
 
 class DomainMap(object):
     """Map representing different coordinate systems.
@@ -208,20 +260,20 @@ class DomainMap(object):
     def __init__(self, grid, mask):
         self.grid = grid
         self.mask = mask
-        ## Constants for conversion between grid- and mask-coordinates
-        self.x_grid2mask = float(mask.nx - 1) / grid.nx
-        self.y_grid2mask = float(mask.ny - 1) / grid.ny
+        # Constants for conversion between grid- and mask-coordinates
+        self.x_grid2mask = (mask.nx - 1) / grid.nx
+        self.y_grid2mask = (mask.ny - 1) / grid.ny
 
         self.x_mask2grid = 1. / self.x_grid2mask
         self.y_mask2grid = 1. / self.y_grid2mask
 
-        self.x_data2grid = grid.nx / grid.width
-        self.y_data2grid = grid.ny / grid.height
+        self.x_data2grid = 1. / grid.dx
+        self.y_data2grid = 1. / grid.dy
 
     def grid2mask(self, xi, yi):
         """Return nearest space in mask-coords from given grid-coords."""
-        return int((xi * self.x_grid2mask) + 0.5), \
-               int((yi * self.y_grid2mask) + 0.5)
+        return (int(xi * self.x_grid2mask + 0.5),
+                int(yi * self.y_grid2mask + 0.5))
 
     def mask2grid(self, xm, ym):
         return xm * self.x_mask2grid, ym * self.y_mask2grid
@@ -229,6 +281,9 @@ class DomainMap(object):
     def data2grid(self, xd, yd):
         return xd * self.x_data2grid, yd * self.y_data2grid
 
+    def grid2data(self, xg, yg):
+        return xg / self.x_data2grid, yg / self.y_data2grid
+
     def start_trajectory(self, xg, yg):
         xm, ym = self.grid2mask(xg, yg)
         self.mask._start_trajectory(xm, ym)
@@ -251,19 +306,25 @@ class Grid(object):
     """Grid of data."""
     def __init__(self, x, y):
 
-        if len(x.shape) == 2:
-            x_row = x[0]
-            assert np.allclose(x_row, x)
+        if x.ndim == 1:
+            pass
+        elif x.ndim == 2:
+            x_row = x[0, :]
+            if not np.allclose(x_row, x):
+                raise ValueError("The rows of 'x' must be equal")
             x = x_row
         else:
-            assert len(x.shape) == 1
+            raise ValueError("'x' can have at maximum 2 dimensions")
 
-        if len(y.shape) == 2:
+        if y.ndim == 1:
+            pass
+        elif y.ndim == 2:
             y_col = y[:, 0]
-            assert np.allclose(y_col, y.T)
+            if not np.allclose(y_col, y.T):
+                raise ValueError("The columns of 'y' must be equal")
             y = y_col
         else:
-            assert len(y.shape) == 1
+            raise ValueError("'y' can have at maximum 2 dimensions")
 
         self.nx = len(x)
         self.ny = len(y)
@@ -277,6 +338,11 @@ class Grid(object):
         self.width = x[-1] - x[0]
         self.height = y[-1] - y[0]
 
+        if not np.allclose(np.diff(x), self.width / (self.nx - 1)):
+            raise ValueError("'x' values must be equally spaced")
+        if not np.allclose(np.diff(y), self.height / (self.ny - 1)):
+            raise ValueError("'y' values must be equally spaced")
+
     @property
     def shape(self):
         return self.ny, self.nx
@@ -298,13 +364,12 @@ class StreamMask(object):
     """
 
     def __init__(self, density):
-        if np.isscalar(density):
-            assert density > 0
-            self.nx = self.ny = int(30 * density)
-        else:
-            assert len(density) == 2
-            self.nx = int(25 * density[0])
-            self.ny = int(25 * density[1])
+        try:
+            self.nx, self.ny = (30 * np.broadcast_to(density, 2)).astype(int)
+        except ValueError:
+            raise ValueError("'density' must be a scalar or be of length 2")
+        if self.nx < 0 or self.ny < 0:
+            raise ValueError("'density' must be positive")
         self._mask = np.zeros((self.ny, self.nx))
         self.shape = self._mask.shape
 
@@ -348,7 +413,7 @@ class TerminateTrajectory(Exception):
 # Integrator definitions
 #========================
 
-def get_integrator(u, v, dmap, minlength):
+def get_integrator(u, v, dmap, minlength, maxlength, integration_direction):
 
     # rescale velocity onto grid-coordinates for integrations.
     u, v = dmap.data2grid(u, v)
@@ -382,14 +447,27 @@ def get_integrator(u, v, dmap, minlength):
         resulting trajectory is None if it is shorter than `minlength`.
         """
 
-        dmap.start_trajectory(x0, y0)
-        sf, xf_traj, yf_traj = _integrate_rk12(x0, y0, dmap, forward_time)
-        dmap.reset_start_point(x0, y0)
-        sb, xb_traj, yb_traj = _integrate_rk12(x0, y0, dmap, backward_time)
-        # combine forward and backward trajectories
-        stotal = sf + sb
-        x_traj = xb_traj[::-1] + xf_traj[1:]
-        y_traj = yb_traj[::-1] + yf_traj[1:]
+        stotal, x_traj, y_traj = 0., [], []
+
+        try:
+            dmap.start_trajectory(x0, y0)
+        except InvalidIndexError:
+            return None
+        if integration_direction in ['both', 'backward']:
+            s, xt, yt = _integrate_rk12(x0, y0, dmap, backward_time, maxlength)
+            stotal += s
+            x_traj += xt[::-1]
+            y_traj += yt[::-1]
+
+        if integration_direction in ['both', 'forward']:
+            dmap.reset_start_point(x0, y0)
+            s, xt, yt = _integrate_rk12(x0, y0, dmap, forward_time, maxlength)
+            if len(x_traj) > 0:
+                xt = xt[1:]
+                yt = yt[1:]
+            stotal += s
+            x_traj += xt
+            y_traj += yt
 
         if stotal > minlength:
             return x_traj, y_traj
@@ -400,7 +478,7 @@ def get_integrator(u, v, dmap, minlength):
     return integrate
 
 
-def _integrate_rk12(x0, y0, dmap, f):
+def _integrate_rk12(x0, y0, dmap, f, maxlength):
     """2nd-order Runge-Kutta algorithm with adaptive step size.
 
     This method is also referred to as the improved Euler's method, or Heun's
@@ -423,17 +501,17 @@ def _integrate_rk12(x0, y0, dmap, f):
     solvers in most setups on my machine. I would recommend removing the
     other two to keep things simple.
     """
-    ## This error is below that needed to match the RK4 integrator. It
-    ## is set for visual reasons -- too low and corners start
-    ## appearing ugly and jagged. Can be tuned.
+    # This error is below that needed to match the RK4 integrator. It
+    # is set for visual reasons -- too low and corners start
+    # appearing ugly and jagged. Can be tuned.
     maxerror = 0.003
 
-    ## This limit is important (for all integrators) to avoid the
-    ## trajectory skipping some mask cells. We could relax this
-    ## condition if we use the code which is commented out below to
-    ## increment the location gradually. However, due to the efficient
-    ## nature of the interpolation, this doesn't boost speed by much
-    ## for quite a bit of complexity.
+    # This limit is important (for all integrators) to avoid the
+    # trajectory skipping some mask cells. We could relax this
+    # condition if we use the code which is commented out below to
+    # increment the location gradually. However, due to the efficient
+    # nature of the interpolation, this doesn't boost speed by much
+    # for quite a bit of complexity.
     maxds = min(1. / dmap.mask.nx, 1. / dmap.mask.ny, 0.1)
 
     ds = maxds
@@ -466,7 +544,7 @@ def _integrate_rk12(x0, y0, dmap, f):
 
         nx, ny = dmap.grid.shape
         # Error is normalized to the axes coordinates
-        error = np.sqrt(((dx2 - dx1) / nx) ** 2 + ((dy2 - dy1) / ny) ** 2)
+        error = np.hypot((dx2 - dx1) / nx, (dy2 - dy1) / ny)
 
         # Only save step if within error tolerance
         if error < maxerror:
@@ -476,7 +554,7 @@ def _integrate_rk12(x0, y0, dmap, f):
                 dmap.update_trajectory(xi, yi)
             except InvalidIndexError:
                 break
-            if (stotal + ds) > 2:
+            if stotal + ds > maxlength:
                 break
             stotal += ds
 
@@ -514,27 +592,27 @@ def _euler_step(xf_traj, yf_traj, dmap, f):
 
 
 # Utility functions
-#========================
+# ========================
 
 def interpgrid(a, xi, yi):
     """Fast 2D, linear interpolation on an integer grid"""
 
     Ny, Nx = np.shape(a)
     if isinstance(xi, np.ndarray):
-        x = xi.astype(np.int)
-        y = yi.astype(np.int)
+        x = xi.astype(int)
+        y = yi.astype(int)
         # Check that xn, yn don't exceed max index
         xn = np.clip(x + 1, 0, Nx - 1)
         yn = np.clip(y + 1, 0, Ny - 1)
     else:
-        x = np.int(xi)
-        y = np.int(yi)
+        x = int(xi)
+        y = int(yi)
         # conditional is faster than clipping for integers
-        if x == (Nx - 2):
+        if x == (Nx - 1):
             xn = x
         else:
             xn = x + 1
-        if y == (Ny - 2):
+        if y == (Ny - 1):
             yn = y
         else:
             yn = y + 1
@@ -569,7 +647,6 @@ def _gen_starting_points(shape):
     xlast = nx - 1
     ylast = ny - 1
     x, y = 0, 0
-    i = 0
     direction = 'right'
     for i in range(nx * ny):
 
